#include <cassert>

#include "box.h"

void Box::in(std::istream* source)
{
   // first input size of the box
   double x, y, z;
   *source >> x >> y >> z;
   this->set_extension_global(x, y, z);
   
   for(int d=0; d < 3; d++)
   {
      this->ext_local_min[d] = (double)(this->boxrank[d])
         * this->extension_global[d] / (double)(this->boxes[d]);
      
      this->ext_local_max[d] = (1.0 + this->boxrank[d])
         * this->extension_global[d] / (double)(this->boxes[d]);
   }
   
   size_t num_part_comp = 0;  // number of particles of a certain component
   this->N_global = 0;
   
   *source >> num_part_comp;  // input number of particles of component 0
   while(num_part_comp != 0)
   {
      // create new component
      double sphere_size = 0.0;
      *source >> sphere_size;
      assert(this->halo_size >= sphere_size);
      this->add_component(sphere_size);
      
      // now read all the particles
      for(size_t i = 0; i < num_part_comp; i++)
      {
         // read the coordinates
         double x[3];
         for(int d = 0; d < 3; d++)
         {
            *source >> x[d];
         
            // apply periodic boundary condition
            while(x[d] < 0.0) x[d] += this->extension_global[d];
            while(x[d] > this->extension_global[d]) x[d] -= this->extension_global[d];
         }
         
         // insert new particle, possibly multiple times, again applying periodic boundary condition
         double y[3];
         int off[3];
         for(off[0] = -1; 1 >= off[0]; off[0]++)
            for(off[1] = -1; 1 >= off[1]; off[1]++)
               for(off[2] = -1; 1 >= off[2]; off[2]++)
               {
                  for(int d = 0; d < 3; d++) y[d] = x[d] + off[d]*this->extension_global[d];
                  this->add_particle(this->N_global, sphere_size, y);
               }
         
         this->N_global++;
      }
      
      *source >> num_part_comp;  // input number of particles for next component
   }
}
std::istream& operator>>(std::istream& is, Box& b) {
   b.in(&is);
   return is;
}

void Box::out(std::ostream* target)
{
   *target << this->extension_global[0] << " " << this->extension_global[1] << " " << this->extension_global[2] << "\n";
   
   for(auto i = this->components.begin(); i != this->components.end(); i++)
   {
      *target << "\n" << this->particles[i->second].size() << " " << i->first << "\n";

      for(auto j = this->particles[i->second].begin(); j != this->particles[i->second].end(); j++)
         *target << j->get_coordinate(0) << "\t"
                 << j->get_coordinate(1) << "\t" << j->get_coordinate(2) << "\n";
   }
   *target << "\n0\n";
}
std::ostream& operator<<(std::ostream& os, Box& b) {
   b.out(&os);
   return os;
}

// count the number of overlaps between pairs of spheres
//
// normal overlaps count as 1, if the inner shielding is broken it counts as 8
//
long Box::count_overlaps()
{
   long overlaps = 0;
   
   // iterate over pairs of components A and B
   for(auto A = this->components.begin(); A != this->components.end(); A++)
      for(auto B = A; B != this->components.end(); B++)
      {
         if(A->second == B->second)  // same component: iterate over pairs of particles i and j
            for(auto i = this->particles[A->second].begin(); std::next(i) != this->particles[A->second].end(); i++)
               for(auto j = std::next(i); j != this->particles[B->second].end(); j++)
               {
                  if(i->get_halo() && j->get_halo()) continue;
                  if(i->get_halo() && (j->get_particle_id() >= i->get_particle_id())) continue;
                  if(j->get_halo() && (i->get_particle_id() >= j->get_particle_id())) continue;
                  
                  overlaps += i->check_overlap(&(*j));
               }

         else  // different components: iterate over pairs of particles i and j
            for(auto i = this->particles[A->second].begin(); i != this->particles[A->second].end(); i++)
               for(auto j = this->particles[B->second].begin(); j != this->particles[B->second].end(); j++)
               {
                  if(i->get_halo() && j->get_halo()) continue;
                  if(i->get_halo() && (j->get_particle_id() >= i->get_particle_id())) continue;
                  if(j->get_halo() && (i->get_particle_id() >= j->get_particle_id())) continue;
                  
                  overlaps += i->check_overlap(&(*j));
               }
      }
   
   return overlaps;
}

// create component, return cid
int Box::add_component(double sphere_size)
{
   int cid = this->particles.size();
   this->components.insert({sphere_size, cid});
   this->particles.push_back(std::vector<Sphere>());
   return cid;
}
   
// create particle, return cid, or return -1 if particle was not added
int Box::add_particle(size_t pid, double sphere_size, double q[3])
{
   bool is_outside_core = false;
   for(int d = 0; d < 3; d++)
   {
      if(q[d] < ext_local_min[d] - this->halo_size) return -1;
      if(q[d] > ext_local_max[d] + this->halo_size) return -1;
      
      if(q[d] < ext_local_min[d]) is_outside_core = true;
      if(q[d] > ext_local_max[d]) is_outside_core = true;
   }
   
   int cid = 0;
   auto comp = this->components.find(sphere_size);
   if(comp != this->components.end()) cid = comp->second;
   else cid = this->add_component(sphere_size);
         
   this->particles[cid].push_back(Sphere(pid, sphere_size, q, is_outside_core));
   
   if(is_outside_core) this->N_local_halo++;
   else this->N_local_core++;
   return cid;
}
