#include <cassert>
#include <chrono>
#include <cstdlib>
#include <fstream>
#include <iostream>

#ifdef USE_MPI
/* MPI header */
#include <mpi.h>
#endif

#include "box.h"
#include "sphere.h"

int main(int argc, char** argv)
{
   assert(argc > 2);
   
   auto t0 = std::chrono::high_resolution_clock::now();

   int rank = 0;
   int size = 1;
   int boxrank[3] = {0, 0, 0};
   int boxes[3] = {1, 1, 1};
#ifdef USE_MPI
   MPI_Init(&argc, &argv);
   
   // how many processes are there?
   MPI_Comm_size(MPI_COMM_WORLD, &size); 
   
   // what is the rank of this process?
   MPI_Comm_rank(MPI_COMM_WORLD, &rank); 
   
   if(argc > 3) boxes[0] = std::atoi(argv[3]);  // split in x direction
   if(argc > 4) boxes[1] = std::atoi(argv[4]);  // split in y direction
   if(argc > 5) boxes[2] = std::atoi(argv[5]);  // split in z direction
   
   int remainder = rank;
   boxrank[0] = remainder / (boxes[1] * boxes[2]);
   remainder -= boxrank[0] * boxes[1] * boxes[2];
   boxrank[1] = remainder / boxes[2];
   remainder -= boxrank[1] * boxes[2];
   boxrank[2] = remainder;
   
   
   for(int i = 0; i < rank; i++) MPI_Barrier(MPI_COMM_WORLD);
   std::cout << "\trank  " << rank << "\t\tbox rank  " << boxrank[0] << "  " << boxrank[1] << "  " << boxrank[2] << "\n";
   std::cout.flush();
   for(int i = rank; i < size; i++) MPI_Barrier(MPI_COMM_WORLD);
#endif
   assert(boxes[0] * boxes[1] * boxes[2] == size);
   
   Box b;
   b.set_boxes_global(boxes);
   b.set_box_rank(boxrank);
   b.set_halo(std::atof(argv[2]));  // halo size must be command line parameter
   std::ifstream file_in(argv[1]);
   file_in >> b;
   file_in.close();
   
   auto t1 = std::chrono::high_resolution_clock::now();
   
   int64_t local_overlaps = b.count_overlaps();
   
   int64_t global_overlaps = local_overlaps;
#ifdef USE_MPI
   if(!rank) std::cout << "\n";
   std::cout.flush();
   for(int i = 0; i < rank; i++) MPI_Barrier(MPI_COMM_WORLD);
   std::cout << "\trank  " << rank << "\t\tlocal overlaps: " << local_overlaps << "\n";
   std::cout.flush();
   for(int i = rank; i < size; i++) MPI_Barrier(MPI_COMM_WORLD);
   
   MPI_Reduce(&local_overlaps, &global_overlaps, 1, MPI_INT64_T, MPI_SUM, 0, MPI_COMM_WORLD);
   
   MPI_Finalize();
#endif
   
   auto t2 = std::chrono::high_resolution_clock::now();
   
   if(!rank)
   {
      std::cout << "\n===\nPotential (overlaps):\t" << global_overlaps << "\n";
   
      auto delta_t01 = std::chrono::duration_cast<std::chrono::microseconds>(t1-t0).count();
      auto delta_t12 = std::chrono::duration_cast<std::chrono::microseconds>(t2-t1).count();
      std::cout << "\nTime to read input:\t" << 1.0e-06*delta_t01 << " s\n";
      std::cout << "To compute overlaps:\t" << 1.0e-06*delta_t12 << " s\n";
      std::cout << "===\nOverall runtime:\t"  << 1.0e-06*(delta_t01+delta_t12) << " s\n";
   }
}
